// This file is part of OpenCV project.
// It is subject to the license terms in the LICENSE file found in the top-level directory
// of this distribution and at http://opencv.org/license.html.

#ifndef OPENCV_NDSRVP_HAL_HPP
#define OPENCV_NDSRVP_HAL_HPP

#include <nds_intrinsic.h>

#include "opencv2/core/hal/interface.h"

#include "include/core.hpp"
#include "include/imgproc.hpp"
#include "include/features2d.hpp"

#endif
