#ifndef __OPENCV_PRECOMP_H__
#define __OPENCV_PRECOMP_H__

#ifdef _MSC_VER
#  define _SILENCE_CXX17_C_HEADER_DEPRECATION_WARNING
#  define _SILENCE_CXX17_CODECVT_HEADER_DEPRECATION_WARNING
#endif

#include <vector>
#include <set>
#include <map>
#include <list>
#include <unordered_set>
#include <unordered_map>
#include <stack>

#include "opencv2/core.hpp"
#include "opencv2/core/types.hpp"
#include "opencv2/core/base.hpp"
#include "opencv2/core/utils/logger.hpp"
#include "opencv2/core/utils/trace.hpp"
#include "opencv2/core/ocl.hpp"
#include "opencv2/core/hal/intrin.hpp"
#include "opencv2/ptcloud.hpp"
#include "opencv2/geometry.hpp"
#include "opencv2/imgproc.hpp"

#endif
