// This file is part of OpenCV project.
// It is subject to the license terms in the LICENSE file found in the top-level directory
// of this distribution and at http://opencv.org/license.html.
//
// Copyright (C) 2025, SpaceMIT Inc., all rights reserved.
// Copyright (C) 2025, Institute of Software, Chinese Academy of Sciences.
// Third party copyrights are property of their respective owners.

#ifndef OPENCV_HAL_RVV_FEATURES2D_COMMON_HPP_INCLUDED
#define OPENCV_HAL_RVV_FEATURES2D_COMMON_HPP_INCLUDED

#include <riscv_vector.h>
#include "opencv2/features/hal/interface.h"

namespace cv { namespace rvv_hal { namespace features2d { namespace common {

#if CV_HAL_RVV_1P0_ENABLED

#endif // CV_HAL_RVV_1P0_ENABLED

}}}} // cv::rvv_hal::core::common

#endif // OPENCV_HAL_RVV_CORE_COMMON_HPP_INCLUDED
