// The file is autogenerated
// Update note: execute <opencv_extra>/testdata/dnn/onnx/generate_conformance_list.py <gtest_xml_file>
"test_adagrad",  // Issues::Layer::Can't create layer "onnx_node_output_0!X1_new" of type "ai.onnx.preview.training.Adagrad" in function 'getLayerInstance'
"test_adagrad_multiple",  // ---- same as above ---
"test_adam",  // Issues::Layer::Can't create layer "onnx_node_output_0!X1_new" of type "ai.onnx.preview.training.Adam" in function 'getLayerInstance'
"test_adam_multiple",  // ---- same as above ---
"test_affine_grid_2d_align_corners_expanded",
"test_affine_grid_2d_expanded",
"test_affine_grid_3d_align_corners_expanded",
"test_affine_grid_3d_expanded",
"test_ai_onnx_ml_array_feature_extractor",
"test_ai_onnx_ml_binarizer",
"test_ai_onnx_ml_label_encoder_string_int",
"test_ai_onnx_ml_label_encoder_string_int_no_default",
"test_ai_onnx_ml_label_encoder_tensor_mapping",
"test_ai_onnx_ml_label_encoder_tensor_value_only_mapping",
"test_ai_onnx_ml_tree_ensemble_set_membership",
"test_ai_onnx_ml_tree_ensemble_single_tree",

// autoSize <= INT_MAX && autoSize*outTotal == inpTotal in function 'getOutShape' @ C++ exception with description
//"OpenCV(5.0.0-pre) opencv/modules/dnn/src/layers/reshape2_layer.cpp:110
// expanded graphs are not imported correctly
"test_attention_3d_expanded",
"test_attention_3d_attn_mask_expanded",
"test_attention_3d_diff_heads_sizes_attn_mask_expanded",
"test_attention_3d_scaled_expanded",
"test_attention_3d_gqa_attn_mask_expanded",
"test_attention_3d_diff_heads_sizes_expanded",
"test_attention_3d_transpose_verification_expanded",
"test_attention_3d_softcap_expanded",
"test_attention_3d_gqa_scaled_expanded",
"test_attention_3d_with_past_and_present_qk_matmul_softmax_expanded",
"test_attention_3d_with_past_and_present_qk_matmul_softcap_expanded",
"test_attention_3d_with_past_and_present_qk_matmul_bias_expanded",
"test_attention_3d_with_past_and_present_qk_matmul_expanded",
"test_attention_3d_with_past_and_present_expanded",
"test_attention_3d_gqa_with_past_and_present_expanded",
"test_attention_3d_causal_expanded",
"test_attention_3d_diff_heads_sizes_causal_expanded",
"test_attention_3d_diff_heads_sizes_scaled_expanded",
"test_attention_3d_diff_heads_sizes_softcap",
"test_attention_3d_diff_heads_sizes_softcap_expanded",
"test_attention_3d_diff_heads_with_past_and_present",
"test_attention_3d_diff_heads_with_past_and_present_expanded",
"test_attention_3d_gqa_causal_expanded",
"test_attention_3d_gqa_expanded",
"test_attention_3d_gqa_softcap_expanded",
"test_attention_4d_attn_mask_3d_causal_expanded",
"test_attention_4d_attn_mask_3d_expanded",
"test_attention_4d_causal_expanded",
"test_attention_4d_diff_heads_sizes_attn_mask_expanded",
"test_attention_4d_diff_heads_mask4d_padded_kv_expanded",
"test_attention_4d_attn_mask_4d_causal_expanded",
"test_attention_4d_attn_mask_4d_expanded",

"test_attention_4d_attn_mask_bool_4d_expanded",
"test_attention_4d_attn_mask_bool_expanded",
"test_attention_4d_attn_mask_expanded",
"test_attention_4d_diff_heads_sizes_causal_expanded",
"test_attention_4d_diff_heads_sizes_expanded",
"test_attention_4d_diff_heads_sizes_scaled_expanded",
"test_attention_4d_diff_heads_sizes_softcap_expanded",
"test_attention_4d_expanded",

"test_attention_3d_gqa_with_past_and_present",
"test_attention_3d_with_past_and_present",
"test_attention_3d_with_past_and_present_qk_matmul",
"test_attention_3d_with_past_and_present_qk_matmul_bias",
"test_attention_3d_with_past_and_present_qk_matmul_softcap",
"test_attention_3d_with_past_and_present_qk_matmul_softmax",
"test_attention_4d_diff_heads_mask4d_padded_kv",
"test_attention_4d_diff_heads_with_past_and_present",
"test_attention_4d_diff_heads_with_past_and_present_expanded",
"test_attention_4d_diff_heads_with_past_and_present_mask3d",
"test_attention_4d_diff_heads_with_past_and_present_mask3d_expanded",
"test_attention_4d_diff_heads_with_past_and_present_mask4d",
"test_attention_4d_diff_heads_with_past_and_present_mask4d_expanded",
"test_attention_4d_fp16_expanded",
"test_attention_4d_gqa_attn_mask_expanded",
"test_attention_4d_gqa_causal_expanded",
"test_attention_4d_gqa_expanded",
"test_attention_4d_gqa_scaled_expanded",
"test_attention_4d_gqa_softcap_expanded",

// fixme
"test_attention_4d_fp16",

"test_attention_4d_gqa_with_past_and_present",
"test_attention_4d_gqa_with_past_and_present_expanded",
"test_attention_4d_gqa_with_past_and_present_fp16",
"test_attention_4d_gqa_with_past_and_present_fp16_expanded",

"test_attention_4d_scaled_expanded",
"test_attention_4d_softcap_expanded",
"test_attention_4d_with_past_and_present",
"test_attention_4d_with_past_and_present_expanded",
"test_attention_4d_with_past_and_present_qk_matmul",
"test_attention_4d_with_past_and_present_qk_matmul_bias",
"test_attention_4d_with_past_and_present_qk_matmul_bias_3d_mask",
"test_attention_4d_with_past_and_present_qk_matmul_bias_3d_mask_causal",
"test_attention_4d_with_past_and_present_qk_matmul_bias_3d_mask_causal_expanded",
"test_attention_4d_with_past_and_present_qk_matmul_bias_3d_mask_expanded",
"test_attention_4d_with_past_and_present_qk_matmul_bias_4d_mask",
"test_attention_4d_with_past_and_present_qk_matmul_bias_4d_mask_causal",
"test_attention_4d_with_past_and_present_qk_matmul_bias_4d_mask_causal_expanded",
"test_attention_4d_with_past_and_present_qk_matmul_bias_4d_mask_expanded",
"test_attention_4d_with_past_and_present_qk_matmul_bias_expanded",
"test_attention_4d_with_past_and_present_qk_matmul_expanded",
"test_attention_4d_with_qk_matmul",
"test_attention_4d_with_qk_matmul_bias",
"test_attention_4d_with_qk_matmul_bias_expanded",
"test_attention_4d_with_qk_matmul_expanded",
"test_attention_4d_with_qk_matmul_softcap",
"test_attention_4d_with_qk_matmul_softcap_expanded",
"test_attention_4d_with_qk_matmul_softmax",
"test_attention_4d_with_qk_matmul_softmax_expanded",
"test_averagepool_2d_ceil_last_window_starts_on_pad",
"test_averagepool_2d_dilations",
"test_averagepool_3d_dilations_large_count_include_pad_is_0_ceil_mode_is_False",
"test_averagepool_3d_dilations_large_count_include_pad_is_0_ceil_mode_is_True",
"test_averagepool_3d_dilations_large_count_include_pad_is_1_ceil_mode_is_False",
"test_averagepool_3d_dilations_large_count_include_pad_is_1_ceil_mode_is_True",
"test_averagepool_3d_dilations_small",
"test_basic_convinteger",  // Issues::Layer::Can't create layer "onnx_node_output_0!y" of type "ConvInteger" in function 'getLayerInstance'
"test_basic_deform_conv_with_padding",
"test_basic_deform_conv_without_padding",
"test_batchnorm_epsilon_training_mode",  // ---- same as above ---
"test_batchnorm_example_training_mode",  // ---- same as above ---
"test_bernoulli",  // Issues::Layer::Can't create layer "onnx_node_output_0!y" of type "Bernoulli" in function 'getLayerInstance'
"test_bernoulli_double",  // ---- same as above ---
"test_bernoulli_double_expanded",  // ---- same as above ---
"test_bernoulli_expanded",  // ---- same as above ---
"test_bernoulli_seed",  // ---- same as above ---
"test_bernoulli_seed_expanded",  // ---- same as above ---
"test_cast_FLOAT16_to_FLOAT4E2M1",
"test_cast_FLOAT16_to_FLOAT8E4M3FN",
"test_cast_FLOAT16_to_FLOAT8E4M3FNUZ",
"test_cast_FLOAT16_to_FLOAT8E5M2",
"test_cast_FLOAT16_to_FLOAT8E5M2FNUZ",
"test_cast_FLOAT16_to_INT4",
"test_cast_FLOAT16_to_UINT4",
"test_cast_FLOAT4E2M1_to_FLOAT",
"test_cast_FLOAT4E2M1_to_FLOAT16",
"test_cast_FLOAT8E4M3FNUZ_to_FLOAT",
"test_cast_FLOAT8E4M3FNUZ_to_FLOAT16",
"test_cast_FLOAT8E4M3FN_to_FLOAT",
"test_cast_FLOAT8E4M3FN_to_FLOAT16",
"test_cast_FLOAT8E5M2FNUZ_to_FLOAT",
"test_cast_FLOAT8E5M2FNUZ_to_FLOAT16",
"test_cast_FLOAT8E5M2_to_FLOAT",
"test_cast_FLOAT8E5M2_to_FLOAT16",
"test_cast_FLOAT_to_FLOAT4E2M1",
"test_cast_FLOAT_to_FLOAT8E4M3FN",
"test_cast_FLOAT_to_FLOAT8E4M3FNUZ",
"test_cast_FLOAT_to_FLOAT8E5M2",
"test_cast_FLOAT_to_FLOAT8E5M2FNUZ",
"test_cast_FLOAT_to_INT4",
"test_cast_FLOAT_to_UINT4",
"test_cast_INT4_to_FLOAT",
"test_cast_INT4_to_FLOAT16",
"test_cast_INT4_to_INT8",
"test_cast_UINT4_to_FLOAT",
"test_cast_UINT4_to_FLOAT16",
"test_cast_UINT4_to_UINT8",
"test_cast_e8m0_FLOAT16_to_FLOAT8E8M0",
"test_cast_e8m0_FLOAT8E8M0_to_FLOAT",
"test_cast_e8m0_FLOAT8E8M0_to_FLOAT16",
"test_cast_e8m0_FLOAT_to_FLOAT8E8M0",
"test_cast_no_saturate_FLOAT16_to_FLOAT8E4M3FN",
"test_cast_no_saturate_FLOAT16_to_FLOAT8E4M3FNUZ",
"test_cast_no_saturate_FLOAT16_to_FLOAT8E5M2",
"test_cast_no_saturate_FLOAT16_to_FLOAT8E5M2FNUZ",
"test_cast_no_saturate_FLOAT_to_FLOAT8E4M3FN",
"test_cast_no_saturate_FLOAT_to_FLOAT8E4M3FNUZ",
"test_cast_no_saturate_FLOAT_to_FLOAT8E5M2",
"test_cast_no_saturate_FLOAT_to_FLOAT8E5M2FNUZ",
"test_castlike_FLOAT16_to_FLOAT4E2M1",
"test_castlike_FLOAT16_to_FLOAT4E2M1_expanded",
"test_castlike_FLOAT16_to_FLOAT8E4M3FN",
"test_castlike_FLOAT16_to_FLOAT8E4M3FNUZ",
"test_castlike_FLOAT16_to_FLOAT8E4M3FNUZ_expanded",
"test_castlike_FLOAT16_to_FLOAT8E4M3FN_expanded",
"test_castlike_FLOAT16_to_FLOAT8E5M2",
"test_castlike_FLOAT16_to_FLOAT8E5M2FNUZ",
"test_castlike_FLOAT16_to_FLOAT8E5M2FNUZ_expanded",
"test_castlike_FLOAT16_to_FLOAT8E5M2_expanded",
"test_castlike_FLOAT16_to_INT4",
"test_castlike_FLOAT16_to_INT4_expanded",
"test_castlike_FLOAT16_to_UINT4",
"test_castlike_FLOAT16_to_UINT4_expanded",
"test_castlike_FLOAT4E2M1_to_FLOAT",
"test_castlike_FLOAT4E2M1_to_FLOAT16",
"test_castlike_FLOAT4E2M1_to_FLOAT16_expanded",
"test_castlike_FLOAT4E2M1_to_FLOAT_expanded",
"test_castlike_FLOAT8E4M3FNUZ_to_FLOAT",
"test_castlike_FLOAT8E4M3FNUZ_to_FLOAT16",
"test_castlike_FLOAT8E4M3FNUZ_to_FLOAT16_expanded",
"test_castlike_FLOAT8E4M3FNUZ_to_FLOAT_expanded",
"test_castlike_FLOAT8E4M3FN_to_FLOAT",
"test_castlike_FLOAT8E4M3FN_to_FLOAT16",
"test_castlike_FLOAT8E4M3FN_to_FLOAT16_expanded",
"test_castlike_FLOAT8E4M3FN_to_FLOAT_expanded",
"test_castlike_FLOAT8E5M2FNUZ_to_FLOAT",
"test_castlike_FLOAT8E5M2FNUZ_to_FLOAT16",
"test_castlike_FLOAT8E5M2FNUZ_to_FLOAT16_expanded",
"test_castlike_FLOAT8E5M2FNUZ_to_FLOAT_expanded",
"test_castlike_FLOAT8E5M2_to_FLOAT",
"test_castlike_FLOAT8E5M2_to_FLOAT16",
"test_castlike_FLOAT8E5M2_to_FLOAT16_expanded",
"test_castlike_FLOAT8E5M2_to_FLOAT_expanded",
"test_castlike_FLOAT_to_FLOAT4E2M1",
"test_castlike_FLOAT_to_FLOAT4E2M1_expanded",
"test_castlike_FLOAT_to_FLOAT8E4M3FN",
"test_castlike_FLOAT_to_FLOAT8E4M3FNUZ",
"test_castlike_FLOAT_to_FLOAT8E4M3FNUZ_expanded",
"test_castlike_FLOAT_to_FLOAT8E4M3FN_expanded",
"test_castlike_FLOAT_to_FLOAT8E5M2",
"test_castlike_FLOAT_to_FLOAT8E5M2FNUZ",
"test_castlike_FLOAT_to_FLOAT8E5M2FNUZ_expanded",
"test_castlike_FLOAT_to_FLOAT8E5M2_expanded",
"test_castlike_FLOAT_to_INT4",
"test_castlike_FLOAT_to_INT4_expanded",
"test_castlike_FLOAT_to_STRING",
"test_castlike_FLOAT_to_UINT4",
"test_castlike_FLOAT_to_UINT4_expanded",
"test_castlike_INT4_to_FLOAT",
"test_castlike_INT4_to_FLOAT16",
"test_castlike_INT4_to_FLOAT16_expanded",
"test_castlike_INT4_to_FLOAT_expanded",
"test_castlike_INT4_to_INT8",
"test_castlike_INT4_to_INT8_expanded",
"test_castlike_STRING_to_FLOAT",  // Issues::Layer::Can't create layer "onnx_node_output_0!output" of type "CastLike" in function 'getLayerInstance'
"test_castlike_UINT4_to_FLOAT",
"test_castlike_UINT4_to_FLOAT16",
"test_castlike_UINT4_to_FLOAT16_expanded",
"test_castlike_UINT4_to_FLOAT_expanded",
"test_castlike_UINT4_to_UINT8",
"test_castlike_UINT4_to_UINT8_expanded",
"test_castlike_no_saturate_FLOAT16_to_FLOAT8E4M3FN",
"test_castlike_no_saturate_FLOAT16_to_FLOAT8E4M3FNUZ",
"test_castlike_no_saturate_FLOAT16_to_FLOAT8E4M3FNUZ_expanded",
"test_castlike_no_saturate_FLOAT16_to_FLOAT8E4M3FN_expanded",
"test_castlike_no_saturate_FLOAT16_to_FLOAT8E5M2",
"test_castlike_no_saturate_FLOAT16_to_FLOAT8E5M2FNUZ",
"test_castlike_no_saturate_FLOAT16_to_FLOAT8E5M2FNUZ_expanded",
"test_castlike_no_saturate_FLOAT16_to_FLOAT8E5M2_expanded",
"test_castlike_no_saturate_FLOAT_to_FLOAT8E4M3FN",
"test_castlike_no_saturate_FLOAT_to_FLOAT8E4M3FNUZ",
"test_castlike_no_saturate_FLOAT_to_FLOAT8E5M2",
"test_castlike_no_saturate_FLOAT_to_FLOAT8E5M2FNUZ",
"test_castlike_no_saturate_FLOAT_to_FLOAT8E5M2FNUZ_expanded",
"test_castlike_no_saturate_FLOAT_to_FLOAT8E5M2_expanded",
"test_clip_example_expanded", //wrong output
"test_clip_expanded",
"test_clip_min_greater_than_max",
"test_clip_min_greater_than_max_expanded",
"test_clip_outbounds_expanded",
"test_col2im",
"test_col2im_5d",
"test_col2im_dilations",
"test_col2im_pads",
"test_col2im_strides",
"test_compress_0",  // Issue::Can't create layer "onnx_node_output_0!output" of type "Compress" in function 'getLayerInstance'
"test_compress_1",  // ---- same as above ---
"test_compress_default_axis",  // ---- same as above ---
"test_compress_negative_axis",  // ---- same as above ---
"test_constant_pad_axes", //type mismatch
"test_constant_pad_negative_axes",
"test_constantofshape_int_shape_zero",  // Issue::Parser::Weights are required as inputs
"test_convinteger_with_padding",  // Issues::Layer::Can't create layer "onnx_node_output_0!y" of type "ConvInteger" in function 'getLayerInstance'
"test_convinteger_without_padding",  //Issues::Layer::Can't create layer "onnx_node_output_0!y" of type "ConvInteger" in function 'getLayerInstance'
"test_convtranspose_autopad_same",
"test_deform_conv_with_mask_bias",
"test_deform_conv_with_multiple_offset_groups",
"test_dequantizelinear_e4m3fn",
"test_dequantizelinear_e4m3fn_float16",
"test_dequantizelinear_e4m3fn_zero_point",
"test_dequantizelinear_e5m2",
"test_dequantizelinear_float4e2m1",
"test_dequantizelinear_int16",
"test_dequantizelinear_int4",
"test_dequantizelinear_uint16",
"test_dequantizelinear_uint4",
"test_dft_axis",
"test_dropout_default_mask",  // Issue::cvtest::norm::wrong data type
"test_dropout_default_mask_ratio",  // ---- same as above ---
"test_dynamicquantizelinear",  // Issue:: Unkonwn error
"test_dynamicquantizelinear_expanded",  // ---- same as above ---
"test_dynamicquantizelinear_max_adjusted",  // ---- same as above ---
"test_dynamicquantizelinear_max_adjusted_expanded",  // ---- same as above ---
"test_dynamicquantizelinear_min_adjusted",  // ---- same as above ---
"test_dynamicquantizelinear_min_adjusted_expanded",  // ---- same as above ---
"test_einsum_inner_prod",  // Issue::Output shape does not match with reference
"test_einsum_scalar",
"test_equal_string",
"test_equal_string_broadcast",
"test_gridsample_bicubic",  // ---- same as above ---
"test_gridsample_bicubic_align_corners_0_additional_1",
"test_gridsample_bicubic_align_corners_1_additional_1",
"test_group_normalization_epsilon_expanded",
"test_group_normalization_example_expanded",
"test_identity_opt",  //  23221 illegal hardware instruction
"test_identity_sequence",  // Issue:: Unkonwn error
"test_if_opt",  // Issue::Failed to allocate 17059022683624350 bytes in function 'OutOfMemoryError'
"test_if_seq",  // Issue::typeProto.has_tensor_type() in function 'dumpValueInfoProto'
"test_image_decoder_decode_bmp_rgb",
"test_image_decoder_decode_jpeg2k_rgb",
"test_image_decoder_decode_jpeg_bgr",
"test_image_decoder_decode_jpeg_grayscale",
"test_image_decoder_decode_jpeg_rgb",
"test_image_decoder_decode_png_rgb",
"test_image_decoder_decode_pnm_rgb",
"test_image_decoder_decode_tiff_rgb",
"test_image_decoder_decode_webp_rgb",
"test_l1normalization_axis_0",
"test_l1normalization_axis_1",
"test_l1normalization_axis_last",
"test_l2normalization_axis_0",
"test_l2normalization_axis_1",
"test_loop13_seq",  // Loop with tensor sequences output, not yet supported in OpenCV
"test_loop16_seq_none",  // Loop with optional tensor sequences, not yet supported in OpenCV
"test_lppool_1d_default",
"test_lppool_2d_default",
"test_lppool_2d_dilations",
"test_lppool_2d_pads",
"test_lppool_2d_same_lower",
"test_lppool_2d_same_upper",
"test_lppool_2d_strides",
"test_lppool_3d_default",
"test_matmulinteger",  // Issues::Layer does not exist. Can't create layer "onnx_node_output_0!Y" of type "MatMulInteger" in function 'getLayerInstance'
"test_maxpool_2d_ceil_output_size_reduce_by_one",
"test_melweightmatrix",
"test_momentum",  // Issues::Layer does not exist. Can't create layer "onnx_node_output_0!X1_new" of type "ai.onnx.preview.training.Momentum" in function 'getLayerInstance'
"test_momentum_multiple",  // ---- same as above ---
"test_mvn",  // Issues::Wrong answer
"test_mvn_expanded",  // Issues::Wrong answer
"test_mvn_expanded_ver18",
"test_nesterov_momentum",  // Issues::Layer does not exist (NesterovsAcceleratedGradient) Can't create layer "onnx_node_output_0!X_new" of type "ai.onnx.preview.training.Momentum" in function 'getLayerInstance'
"test_nllloss_NCd1d2_reduction_sum_expanded",
"test_nllloss_NCd1d2d3d4d5_mean_weight_expanded",
"test_optional_get_element",  // Issue::out of memory :: Failed to allocate 1044051907127083008 bytes in function 'OutOfMemoryError'
"test_optional_get_element_optional_sequence",
"test_optional_get_element_optional_tensor",
"test_optional_get_element_sequence",  // ---- same as above ---
"test_optional_get_element_tensor",
"test_optional_has_element",  // Issue::typeProto.has_tensor_type() in function 'populateNet'
"test_optional_has_element_empty",  // ---- same as above ---
"test_optional_has_element_empty_no_input_name_optional_input",
"test_optional_has_element_empty_no_input_name_tensor_input",
"test_optional_has_element_empty_no_input_optional_input",
"test_optional_has_element_empty_no_input_tensor_input",
"test_optional_has_element_empty_optional_input",
"test_optional_has_element_optional_input",
"test_optional_has_element_tensor_input",
"test_prelu_broadcast",  // Issue::Parser:Blob slope not found in const blobs in function 'getBlob' (weights are required as inputs)
"test_prelu_example",  // ---- same as above ---
"test_qlinearmatmul_2D_int8_float16",  // Float output QLinearMatMul variants not supported
"test_qlinearmatmul_2D_int8_float32",
"test_qlinearmatmul_2D_uint8_float16",
"test_qlinearmatmul_2D_uint8_float32",
"test_qlinearmatmul_3D_int8_float16",  // Float output QLinearMatMul variants not supported
"test_qlinearmatmul_3D_int8_float32",
"test_qlinearmatmul_3D_uint8_float16",
"test_qlinearmatmul_3D_uint8_float32",
"test_quantizelinear_blocked_asymmetric",
"test_quantizelinear_blocked_symmetric",
"test_quantizelinear_e4m3fn",
"test_quantizelinear_e5m2",
"test_quantizelinear_float4e2m1",
"test_quantizelinear_int16",
"test_quantizelinear_int4",
"test_quantizelinear_uint16",
"test_quantizelinear_uint4",
"test_range_float_type_positive_delta_expanded",  // ---- Unsupported operations: Loop ---
"test_range_int32_type_negative_delta_expanded",  // ---- same as above ---
"test_regex_full_match_basic",
"test_regex_full_match_email_domain",
"test_regex_full_match_empty",
"test_reshape_allowzero_reordered",  // incompatible type of input tensor #0 'data': CV_8UC1 given, CV_32FC1 expected in function 'setGraphInput'
"test_resize_downsample_scales_cubic_align_corners",  // ---- same as above ---
"test_resize_downsample_scales_cubic_antialias",
"test_resize_downsample_scales_linear_align_corners",  // ---- same as above ---
"test_resize_downsample_scales_linear_antialias", //incorrect output
"test_resize_downsample_scales_linear_half_pixel_symmetric",
"test_resize_downsample_sizes_cubic_antialias",
"test_resize_downsample_sizes_linear_antialias",
"test_resize_downsample_sizes_nearest_not_larger",
"test_resize_downsample_sizes_nearest_not_smaller",
"test_resize_tf_crop_and_resize_axes_2_3",
"test_resize_tf_crop_and_resize_axes_3_2",
"test_resize_tf_crop_and_resize_extrapolation_value",
"test_resize_upsample_scales_linear_half_pixel_symmetric", //shape mismatch
"test_resize_upsample_scales_nearest_axes_3_2",
"test_resize_upsample_sizes_nearest_axes_3_2",
"test_resize_upsample_sizes_nearest_not_larger",
"test_resize_upsample_sizes_nearest_not_smaller",
"test_reversesequence_batch",  // Issue:: Parser: Can't create layer "onnx_node_output_0!y" of type "ReverseSequence" in function 'getLayerInstance'
"test_reversesequence_time",  // ---- same as above ---
"test_rms_normalization_2d_axis0_expanded",
"test_rms_normalization_2d_axis1_expanded",
"test_rms_normalization_2d_axis_negative_1_expanded",
"test_rms_normalization_2d_axis_negative_2_expanded",
"test_rms_normalization_3d_axis0_epsilon_expanded",
"test_rms_normalization_3d_axis1_epsilon_expanded",
"test_rms_normalization_3d_axis2_epsilon_expanded",
"test_rms_normalization_3d_axis_negative_1_epsilon_expanded",
"test_rms_normalization_3d_axis_negative_2_epsilon_expanded",
"test_rms_normalization_3d_axis_negative_3_epsilon_expanded",
"test_rms_normalization_4d_axis0_expanded",
"test_rms_normalization_4d_axis1_expanded",
"test_rms_normalization_4d_axis2_expanded",
"test_rms_normalization_4d_axis3_expanded",
"test_rms_normalization_4d_axis_negative_1_expanded",
"test_rms_normalization_4d_axis_negative_2_expanded",
"test_rms_normalization_4d_axis_negative_3_expanded",
"test_rms_normalization_4d_axis_negative_4_expanded",
"test_rms_normalization_default_axis_expanded",
"test_rnn_seq_length",  // Issue:: Parser:  Can't create layer "onnx_node_output_1!Y_h" of type "RNN" in function 'getLayerInstance'
"test_scan9_sum",  // Issue:: Parser: 'Graph' is not supported in function 'getLayerParams'
"test_scan_sum", // ---- same as above ---
"test_sequence_insert_at_back",  // Issue:: Parser:  typeProto.has_tensor_type() in function 'populateNet'
"test_sequence_insert_at_front",  // ---- same as above ---
"test_sequence_map_add_1_sequence_1_tensor",
"test_sequence_map_add_1_sequence_1_tensor_expanded",
"test_sequence_map_add_2_sequences",
"test_sequence_map_add_2_sequences_expanded",
"test_sequence_map_extract_shapes",
"test_sequence_map_extract_shapes_expanded",
"test_sequence_map_identity_1_sequence",
"test_sequence_map_identity_1_sequence_1_tensor",
"test_sequence_map_identity_1_sequence_1_tensor_expanded",
"test_sequence_map_identity_1_sequence_expanded",
"test_sequence_map_identity_2_sequences",
"test_sequence_map_identity_2_sequences_expanded",
"test_shape_start_greater_than_end",
"test_simple_rnn_batchwise",  // Issue:: Parser: Can't create layer "onnx_node_output_1!Y_h" of type "RNN" in function 'getLayerInstance'
"test_simple_rnn_defaults",  // ---- same as above ---
"test_simple_rnn_with_initial_bias",  // ---- same as above ---
"test_slice_start_out_of_bounds",
"test_split_1d_uneven_split_opset18", //type mismatch
"test_split_2d_uneven_split_opset18",
"test_split_equal_parts_1d_opset13",
"test_split_equal_parts_1d_opset18",
"test_split_equal_parts_default_axis_opset13",
"test_split_equal_parts_default_axis_opset18",
"test_split_to_sequence_1",
"test_split_to_sequence_2",
"test_split_to_sequence_nokeepdims",
"test_split_zero_size_splits",  // ---- incompatible type of input tensor #0 'input': CV_8UC1 given, CV_32FC1 expected in function 'setGraphInput' ---
"test_split_zero_size_splits_opset13", // type mismatch
"test_split_zero_size_splits_opset18", // type mismatch
"test_stft",
"test_stft_with_window",
"test_string_concat",
"test_string_concat_broadcasting",
"test_string_concat_empty_string",
"test_string_concat_utf8",
"test_string_concat_zero_dimensional",
"test_string_split_basic",
"test_string_split_consecutive_delimiters",
"test_string_split_empty_string_delimiter",
"test_string_split_empty_tensor",
"test_string_split_maxsplit",
"test_string_split_no_delimiter",
"test_strnormalizer_export_monday_casesensintive_lower",  // 'Strings' (1) are not supported in function 'getLayerParams'
"test_strnormalizer_export_monday_casesensintive_nochangecase",  // ---- same as above ---
"test_strnormalizer_export_monday_casesensintive_upper",  // ---- same as above ---
"test_strnormalizer_export_monday_empty_output",  // ---- same as above ---
"test_strnormalizer_export_monday_insensintive_upper_twodim",  // ---- same as above ---
"test_strnormalizer_nostopwords_nochangecase",  // Issue:: Parser: Can't create layer "onnx_node_output_0!y" of type "StringNormalizer" in function 'getLayerInstance'
"test_swish_expanded",
"test_tensorscatter",
"test_tensorscatter_3d",
"test_tensorscatter_circular",
"test_tfidfvectorizer_tf_batch_onlybigrams_skip0",  // Issue:: Parser: Can't create layer "onnx_node_output_0!Y" of type "TfIdfVectorizer" in function 'getLayerInstance'
"test_tfidfvectorizer_tf_batch_onlybigrams_skip5",  // ---- same as above ---
"test_tfidfvectorizer_tf_batch_uniandbigrams_skip5",  // ---- same as above ---
"test_tfidfvectorizer_tf_only_bigrams_skip0",  // ---- same as above ---
"test_tfidfvectorizer_tf_onlybigrams_levelempty",  // ---- same as above ---
"test_tfidfvectorizer_tf_onlybigrams_skip5",  // ---- same as above ---
"test_tfidfvectorizer_tf_uniandbigrams_skip5",  // Issue:: Parser: Can't create layer "onnx_node_output_0!Y" of type "TfIdfVectorizer" in function 'getLayerInstance'
"test_training_dropout",  // Issue::cvtest::norm::wrong data type
"test_training_dropout_default",  // ---- same as above --- type mismatch
"test_training_dropout_default_mask",  // ---- same as above ---
"test_training_dropout_mask",  // ---- same as above ---
"test_training_dropout_zero_ratio_mask",  // ---- same as above ---
"test_tril_zero",  // ---- same as above --- type mismatch
"test_triu_zero",  // ---- same as above --- type mismatch
"test_unique_length_1", //incorrect output
