"test_if",
"test_top_k", // Issue:: K being input is not compatible with the current engine
"test_top_k_negative_axis",  //  same as above
"test_top_k_smallest",  //  same as above
"test_expand_dim_changed",
"test_expand_dim_unchanged",
"test_gemm_all_attributes",
"test_gemm_alpha",
"test_gemm_beta",
"test_gemm_default_scalar_bias",
"test_gemm_default_single_elem_vector_bias",
"test_gemm_default_vector_bias",
"test_gemm_default_zero_bias",
"test_gemm_transposeA",
"test_gemm_transposeB",
"test_range_float_type_positive_delta",
"test_range_int32_type_negative_delta",
"test_reshape_extended_dims",
"test_reshape_negative_dim",
"test_reshape_negative_extended_dims",
"test_reshape_one_dim",
"test_reshape_reduced_dims",
"test_reshape_reordered_all_dims",
"test_reshape_reordered_last_dims",
"test_reshape_zero_and_negative_dim",
"test_reshape_zero_dim",
"test_shape",
"test_shape_clip_end",
"test_shape_clip_start",
"test_shape_end_1",
"test_shape_end_negative_1",
"test_shape_example",
"test_shape_start_1",
"test_shape_start_1_end_2",
"test_shape_start_1_end_negative_1",
"test_shape_start_negative_1",
"test_slice",
"test_slice_default_axes",
"test_slice_default_steps",
"test_slice_end_out_of_bounds",
"test_slice_neg",
"test_slice_neg_steps",
"test_slice_negative_axes",
"test_split_variable_parts_1d",
"test_split_variable_parts_2d",
"test_split_variable_parts_default_axis",
"test_squeeze",
"test_squeeze_negative_axes",
"test_tile",
"test_tile_precomputed",
"test_unsqueeze_axis_0",
"test_unsqueeze_axis_1",
"test_unsqueeze_axis_2",
"test_unsqueeze_negative_axes",
"test_unsqueeze_three_axes",
"test_unsqueeze_two_axes",
"test_unsqueeze_unsorted_axes",
"test_clip",
"test_clip_default_inbounds",
"test_clip_default_int8_inbounds",
"test_clip_default_int8_max",
"test_clip_default_int8_min",
"test_clip_default_max",
"test_clip_default_min",
"test_clip_example",
"test_clip_inbounds",
"test_clip_outbounds",
"test_clip_splitbounds",
"test_size",
"test_size_example",
"test_mean_example",
"test_mean_one_input",
"test_mean_two_inputs",
"test_isnan",
"test_isinf",
"test_isinf_negative",
"test_isinf_positive",
"test_tril",
"test_tril_neg",
"test_tril_one_row_neg",
"test_tril_out_neg",
"test_tril_out_pos",
"test_tril_pos",
"test_tril_square",
"test_tril_square_neg",
"test_triu",
"test_triu_neg",
"test_triu_one_row",
"test_triu_out_neg_out",
"test_triu_out_pos",
"test_triu_pos",
"test_triu_square",
"test_triu_square_neg",
"test_det_2d",
"test_det_nd",
"test_max_int16",
"test_max_uint16",
"test_max_uint32",
"test_max_uint64",
"test_min_int16",
"test_min_uint16",
"test_min_uint32",
"test_min_uint64",
"test_mod_mixed_sign_int16",
"test_mod_uint16",
"test_mod_uint32",
"test_mod_uint64",
"test_bitshift_left_uint16",
"test_bitshift_left_uint32",
"test_bitshift_left_uint64",
"test_bitshift_left_uint8",
"test_bitshift_right_uint16",
"test_bitshift_right_uint32",
"test_bitshift_right_uint64",
"test_bitshift_right_uint8",
"test_gridsample",
"test_gridsample_aligncorners_true",
"test_gridsample_bilinear",
"test_gridsample_border_padding",
"test_gridsample_reflection_padding",
"test_gridsample_zeros_padding",
"test_gridsample_nearest",
"test_edge_pad",
"test_lstm_batchwise",
"test_lstm_defaults",
"test_lstm_with_initial_bias",
"test_lstm_with_peepholes",
"test_pow_types_float",
"test_pow_types_float32_int32",
"test_pow_types_float32_int64",
"test_pow_types_float32_uint32",
"test_pow_types_float32_uint64",
"test_pow_types_int",
"test_pow_types_int32_float32",
"test_pow_types_int32_int32",
"test_pow_types_int64_float32",
"test_reflect_pad",
"test_constant",
"test_constant_pad",
"test_constantofshape_float_ones",
"test_constantofshape_int_zeros",
"test_nonzero_example",
"test_unique_not_sorted_without_axis",
"test_unique_sorted_with_axis",
"test_unique_sorted_with_axis_3d",
"test_unique_sorted_with_negative_axis",
"test_unique_sorted_without_axis",
"test_resize_downsample_scales_nearest",
"test_resize_downsample_sizes_cubic",
"test_resize_downsample_sizes_nearest",
"test_resize_upsample_scales_cubic",
"test_resize_upsample_scales_cubic_A_n0p5_exclude_outside",
"test_resize_upsample_scales_cubic_align_corners",
"test_resize_upsample_scales_cubic_asymmetric",
"test_resize_upsample_scales_linear",
"test_resize_upsample_scales_linear_align_corners",
"test_resize_upsample_scales_nearest",
"test_resize_upsample_sizes_cubic",
"test_resize_upsample_sizes_nearest",
"test_resize_upsample_sizes_nearest_ceil_half_pixel",
"test_resize_upsample_sizes_nearest_floor_align_corners",
"test_resize_upsample_sizes_nearest_round_prefer_ceil_asymmetric",
"test_resize_downsample_scales_cubic",
"test_resize_downsample_scales_cubic_A_n0p5_exclude_outside",
"test_resize_downsample_scales_linear",
"test_resize_downsample_sizes_linear_pytorch_half_pixel",
"test_resize_downsample_sizes_nearest_tf_half_pixel_for_nn",
"test_resize_tf_crop_and_resize",
"test_nonmaxsuppression_center_point_box_format",
"test_nonmaxsuppression_flipped_coordinates",
"test_nonmaxsuppression_identical_boxes",
"test_nonmaxsuppression_limit_output_size",
"test_nonmaxsuppression_single_box",
"test_nonmaxsuppression_suppress_by_IOU",
"test_nonmaxsuppression_suppress_by_IOU_and_scores",
"test_nonmaxsuppression_two_batches",
"test_nonmaxsuppression_two_classes",
"test_add_int16",
"test_add_int8",
"test_add_uint16",
"test_add_uint32",
"test_add_uint64",
"test_clip_default_inbounds_expanded",
"test_clip_default_int8_inbounds_expanded",
"test_clip_default_int8_max_expanded",
"test_clip_default_int8_min_expanded",
"test_clip_default_max_expanded",
"test_clip_default_min_expanded",
"test_clip_inbounds_expanded",
"test_clip_splitbounds_expanded",
"test_equal_int16",
"test_equal_int8",
"test_equal_uint16",
"test_equal_uint32",
"test_equal_uint64",
"test_equal_uint8",
"test_isinf_float16",
"test_isnan_float16",
"test_logsoftmax_axis_0_expanded_ver18",
"test_logsoftmax_axis_1_expanded_ver18",
"test_logsoftmax_axis_2_expanded_ver18",
"test_logsoftmax_default_axis_expanded_ver18",
"test_logsoftmax_example_1_expanded_ver18",
"test_logsoftmax_large_number_expanded_ver18",
"test_logsoftmax_negative_axis_expanded_ver18",
"test_mul_int16",
"test_mul_int8",
"test_mul_uint16",
"test_mul_uint32",
"test_mul_uint64",
"test_softmax_axis_0_expanded_ver18",
"test_softmax_axis_1_expanded_ver18",
"test_softmax_axis_2_expanded_ver18",
"test_softmax_default_axis_expanded_ver18",
"test_softmax_example_expanded_ver18",
"test_softmax_large_number_expanded_ver18",
"test_softmax_negative_axis_expanded_ver18",
"test_swish",
"test_greater_equal_int16",
"test_greater_equal_int16_expanded",
"test_greater_equal_int8",
"test_greater_equal_int8_expanded",
"test_greater_equal_uint16",
"test_greater_equal_uint16_expanded",
"test_greater_equal_uint32",
"test_greater_equal_uint32_expanded",
"test_greater_equal_uint64",
"test_greater_equal_uint64_expanded",
"test_greater_equal_uint8",
"test_greater_equal_uint8_expanded",
"test_greater_int16",
"test_greater_int8",
"test_greater_uint16",
"test_greater_uint32",
"test_greater_uint64",
"test_greater_uint8",
"test_less_equal_int16",
"test_less_equal_int16_expanded",
"test_less_equal_int8",
"test_less_equal_int8_expanded",
"test_less_equal_uint16",
"test_less_equal_uint16_expanded",
"test_less_equal_uint32",
"test_less_equal_uint32_expanded",
"test_less_equal_uint64",
"test_less_equal_uint64_expanded",
"test_less_equal_uint8",
"test_less_equal_uint8_expanded",
"test_less_int16",
"test_less_int8",
"test_less_uint16",
"test_less_uint32",
"test_less_uint64",
"test_less_uint8",
"test_lpnormalization_default",
"test_resize_upsample_scales_nearest_axes_2_3",
"test_resize_upsample_sizes_nearest_axes_2_3",
"test_split_equal_parts_2d_opset13",
"test_split_variable_parts_1d_opset13",
"test_split_variable_parts_1d_opset18",
"test_split_variable_parts_2d_opset13",
"test_split_variable_parts_2d_opset18",
"test_split_variable_parts_default_axis_opset13",
"test_split_variable_parts_default_axis_opset18",
"test_sub_int16",
"test_sub_int8",
"test_sub_uint16",
"test_sub_uint32",
"test_sub_uint64",
"test_top_k_same_values", //type mismatch
"test_top_k_same_values_2d",
"test_top_k_same_values_largest",
"test_top_k_uint64",
"test_training_dropout_zero_ratio",  // ---- same as above ---
"test_wrap_pad", // type mismatch
"test_div_int16",
"test_div_uint8",
"test_div_int8",
"test_div_uint16",
"test_div_uint32",
"test_div_uint64",
"test_cumsum_1d_int32_exclusive",
"test_cumsum_2d_int32",
"test_cast_BFLOAT16_to_FLOAT",
"test_cast_DOUBLE_to_FLOAT16",
"test_cast_FLOAT16_to_DOUBLE",
"test_cast_FLOAT16_to_FLOAT",
"test_cast_FLOAT_to_BFLOAT16",
"test_cast_FLOAT_to_DOUBLE",
"test_cast_FLOAT_to_FLOAT16",
"test_castlike_BFLOAT16_to_FLOAT",
"test_castlike_BFLOAT16_to_FLOAT_expanded",
"test_castlike_DOUBLE_to_FLOAT",
"test_castlike_DOUBLE_to_FLOAT16",
"test_castlike_DOUBLE_to_FLOAT16_expanded",
"test_castlike_DOUBLE_to_FLOAT_expanded",
"test_castlike_FLOAT16_to_DOUBLE",
"test_castlike_FLOAT16_to_DOUBLE_expanded",
"test_castlike_FLOAT16_to_FLOAT",
"test_castlike_FLOAT16_to_FLOAT_expanded",
"test_castlike_FLOAT_to_BFLOAT16",
"test_castlike_FLOAT_to_BFLOAT16_expanded",
"test_castlike_FLOAT_to_DOUBLE",
"test_castlike_FLOAT_to_DOUBLE_expanded",
"test_castlike_FLOAT_to_FLOAT16",
"test_castlike_FLOAT_to_FLOAT16_expanded",
"test_gelu_default_1_expanded",
"test_gelu_default_2_expanded",
"test_gelu_tanh_1_expanded",
"test_gelu_tanh_2_expanded",
"test_bitwise_and_i16_3d",
"test_bitwise_and_i32_2d",
"test_bitwise_and_ui64_bcast_3v1d",
"test_bitwise_and_ui8_bcast_4v3d",
"test_bitwise_not_2d",
"test_bitwise_not_3d",
"test_bitwise_not_4d",
"test_bitwise_or_i16_4d",
"test_bitwise_or_i32_2d",
"test_bitwise_or_ui64_bcast_3v1d",
"test_bitwise_or_ui8_bcast_4v3d",
"test_bitwise_xor_i16_3d",
"test_bitwise_xor_i32_2d",
"test_bitwise_xor_ui64_bcast_3v1d",
"test_bitwise_xor_ui8_bcast_4v3d",
"test_reduce_sum_default_axes_keepdims_example",
"test_reduce_sum_do_not_keepdims_example",
"test_reduce_sum_do_not_keepdims_random",
"test_reduce_sum_empty_axes_input_noop_example",
"test_reduce_sum_empty_axes_input_noop_random",
"test_reduce_sum_keepdims_example",
"test_reduce_sum_keepdims_random",
"test_reduce_sum_negative_axes_keepdims_example",
"test_reduce_sum_negative_axes_keepdims_random",
"test_reduce_sum_default_axes_keepdims_random",
"test_reduce_l1_default_axes_keepdims_example_expanded",
"test_reduce_l1_default_axes_keepdims_random_expanded",
"test_reduce_l1_do_not_keepdims_example_expanded",
"test_reduce_l1_do_not_keepdims_random_expanded",
"test_reduce_l1_keep_dims_example_expanded",
"test_reduce_l1_keep_dims_random_expanded",
"test_reduce_l1_negative_axes_keep_dims_example_expanded",
"test_reduce_l1_negative_axes_keep_dims_random_expanded",
"test_reduce_log_sum_asc_axes_expanded",
"test_reduce_log_sum_default_expanded",
"test_reduce_log_sum_desc_axes_expanded",
"test_reduce_log_sum_negative_axes_expanded",
"test_reduce_max_bool_inputs",
"test_reduce_min_bool_inputs",
"test_reduce_sum_square_default_axes_keepdims_example_expanded",
"test_reduce_sum_square_default_axes_keepdims_random_expanded",
"test_reduce_sum_square_do_not_keepdims_example_expanded",
"test_reduce_sum_square_do_not_keepdims_random_expanded",
"test_reduce_sum_square_keepdims_example_expanded",
"test_reduce_sum_square_keepdims_random_expanded",
"test_reduce_sum_square_negative_axes_keepdims_example_expanded",
"test_reduce_sum_square_negative_axes_keepdims_random_expanded",
"test_reduce_sum_empty_axes_input_noop",
"test_reduce_l1_default_axes_keepdims_example",
"test_reduce_l1_default_axes_keepdims_random",
"test_reduce_l1_do_not_keepdims_example",
"test_reduce_l1_do_not_keepdims_random",
"test_reduce_l1_keep_dims_example",
"test_reduce_l1_keep_dims_random",
"test_reduce_l1_negative_axes_keep_dims_example",
"test_reduce_l1_negative_axes_keep_dims_random",
"test_reduce_l2_default_axes_keepdims_example",
"test_reduce_l2_default_axes_keepdims_example_expanded",
"test_reduce_l2_default_axes_keepdims_random",
"test_reduce_l2_default_axes_keepdims_random_expanded",
"test_reduce_l2_do_not_keepdims_example",
"test_reduce_l2_do_not_keepdims_example_expanded",
"test_reduce_l2_do_not_keepdims_random",
"test_reduce_l2_do_not_keepdims_random_expanded",
"test_reduce_l2_keep_dims_example",
"test_reduce_l2_keep_dims_example_expanded",
"test_reduce_l2_keep_dims_random",
"test_reduce_l2_keep_dims_random_expanded",
"test_reduce_l2_negative_axes_keep_dims_example",
"test_reduce_l2_negative_axes_keep_dims_example_expanded",
"test_reduce_l2_negative_axes_keep_dims_random",
"test_reduce_l2_negative_axes_keep_dims_random_expanded",
"test_reduce_log_sum_asc_axes",
"test_reduce_log_sum_default",
"test_reduce_log_sum_desc_axes",
"test_reduce_log_sum_exp_default_axes_keepdims_example",
"test_reduce_log_sum_exp_default_axes_keepdims_example_expanded",
"test_reduce_log_sum_exp_default_axes_keepdims_random",
"test_reduce_log_sum_exp_default_axes_keepdims_random_expanded",
"test_reduce_log_sum_exp_do_not_keepdims_example",
"test_reduce_log_sum_exp_do_not_keepdims_example_expanded",
"test_reduce_log_sum_exp_do_not_keepdims_random",
"test_reduce_log_sum_exp_do_not_keepdims_random_expanded",
"test_reduce_log_sum_exp_keepdims_example",
"test_reduce_log_sum_exp_keepdims_example_expanded",
"test_reduce_log_sum_exp_keepdims_random",
"test_reduce_log_sum_exp_keepdims_random_expanded",
"test_reduce_log_sum_exp_negative_axes_keepdims_example",
"test_reduce_log_sum_exp_negative_axes_keepdims_example_expanded",
"test_reduce_log_sum_exp_negative_axes_keepdims_random",
"test_reduce_log_sum_exp_negative_axes_keepdims_random_expanded",
"test_reduce_log_sum_negative_axes",
"test_reduce_max_do_not_keepdims_example",
"test_reduce_max_do_not_keepdims_random",
"test_reduce_max_keepdims_example",
"test_reduce_max_keepdims_random",
"test_reduce_max_negative_axes_keepdims_example",
"test_reduce_max_negative_axes_keepdims_random",
"test_reduce_mean_default_axes_keepdims_example",
"test_reduce_mean_default_axes_keepdims_random",
"test_reduce_mean_do_not_keepdims_example",
"test_reduce_mean_do_not_keepdims_random",
"test_reduce_mean_keepdims_example",
"test_reduce_mean_keepdims_random",
"test_reduce_mean_negative_axes_keepdims_example",
"test_reduce_mean_negative_axes_keepdims_random",
"test_reduce_min_do_not_keepdims_example",
"test_reduce_min_do_not_keepdims_random",
"test_reduce_min_keepdims_example",
"test_reduce_min_keepdims_random",
"test_reduce_min_negative_axes_keepdims_example",
"test_reduce_min_negative_axes_keepdims_random",
"test_reduce_prod_do_not_keepdims_example",
"test_reduce_prod_do_not_keepdims_random",
"test_reduce_prod_keepdims_example",
"test_reduce_prod_keepdims_random",
"test_reduce_prod_negative_axes_keepdims_example",
"test_reduce_prod_negative_axes_keepdims_random",
"test_reduce_sum_square_default_axes_keepdims_example",
"test_reduce_sum_square_default_axes_keepdims_random",
"test_reduce_sum_square_do_not_keepdims_example",
"test_reduce_sum_square_do_not_keepdims_random",
"test_reduce_sum_square_keepdims_example",
"test_reduce_sum_square_keepdims_random",
"test_reduce_sum_square_negative_axes_keepdims_example",
"test_reduce_sum_square_negative_axes_keepdims_random",
"test_elu_default_expanded_ver18",
"test_elu_example_expanded_ver18",
"test_elu_expanded_ver18",
"test_thresholdedrelu_default_expanded_ver18",
"test_thresholdedrelu_example_expanded_ver18",
"test_thresholdedrelu_expanded_ver18",
"test_selu_default_expanded_ver18",
"test_selu_example_expanded_ver18",
"test_selu_expanded_ver18",
"test_hardsigmoid_default_expanded_ver18",
"test_hardsigmoid_example_expanded_ver18",
"test_hardsigmoid_expanded_ver18",
"test_softplus_example_expanded_ver18",
"test_softplus_expanded_ver18",
"test_softsign_example_expanded_ver18",
"test_softsign_expanded_ver18",
"test_shrink_hard_expanded_ver18",
"test_shrink_soft_expanded_ver18",
"test_relu_expanded_ver18",
"test_prelu_broadcast_expanded",
"test_prelu_example_expanded",
"test_leakyrelu_default_expanded",
"test_leakyrelu_example_expanded",
"test_leakyrelu_expanded",
"test_sce_NCd1_mean_weight_negative_ii",
"test_sce_NCd1_mean_weight_negative_ii_expanded",
"test_sce_NCd1_mean_weight_negative_ii_log_prob",
"test_sce_NCd1_mean_weight_negative_ii_log_prob_expanded",
"test_sce_NCd1d2d3_none_no_weight_negative_ii",
"test_sce_NCd1d2d3_none_no_weight_negative_ii_expanded",
"test_sce_NCd1d2d3_none_no_weight_negative_ii_log_prob",
"test_sce_NCd1d2d3_none_no_weight_negative_ii_log_prob_expanded",
"test_sce_NCd1d2d3_sum_weight_high_ii",
"test_sce_NCd1d2d3_sum_weight_high_ii_expanded",
"test_sce_NCd1d2d3_sum_weight_high_ii_log_prob",
"test_sce_NCd1d2d3_sum_weight_high_ii_log_prob_expanded",
"test_sce_NCd1d2d3d4d5_mean_weight",
"test_sce_NCd1d2d3d4d5_mean_weight_expanded",
"test_sce_NCd1d2d3d4d5_mean_weight_log_prob",
"test_sce_NCd1d2d3d4d5_mean_weight_log_prob_expanded",
"test_sce_NCd1d2d3d4d5_none_no_weight",
"test_sce_NCd1d2d3d4d5_none_no_weight_expanded",
"test_sce_NCd1d2d3d4d5_none_no_weight_log_prob",
"test_sce_NCd1d2d3d4d5_none_no_weight_log_prob_expanded",
"test_sce_mean",
"test_sce_mean_3d",
"test_sce_mean_3d_expanded",
"test_sce_mean_3d_log_prob",
"test_sce_mean_3d_log_prob_expanded",
"test_sce_mean_expanded",
"test_sce_mean_log_prob",
"test_sce_mean_log_prob_expanded",
"test_sce_mean_no_weight_ii",
"test_sce_mean_no_weight_ii_3d",
"test_sce_mean_no_weight_ii_3d_expanded",
"test_sce_mean_no_weight_ii_3d_log_prob",
"test_sce_mean_no_weight_ii_3d_log_prob_expanded",
"test_sce_mean_no_weight_ii_4d",
"test_sce_mean_no_weight_ii_4d_expanded",
"test_sce_mean_no_weight_ii_4d_log_prob",
"test_sce_mean_no_weight_ii_4d_log_prob_expanded",
"test_sce_mean_no_weight_ii_expanded",
"test_sce_mean_no_weight_ii_log_prob",
"test_sce_mean_no_weight_ii_log_prob_expanded",
"test_sce_mean_weight",
"test_sce_mean_weight_expanded",
"test_sce_mean_weight_ii",
"test_sce_mean_weight_ii_3d",
"test_sce_mean_weight_ii_3d_expanded",
"test_sce_mean_weight_ii_3d_log_prob",
"test_sce_mean_weight_ii_3d_log_prob_expanded",
"test_sce_mean_weight_ii_4d",
"test_sce_mean_weight_ii_4d_expanded",
"test_sce_mean_weight_ii_4d_log_prob",
"test_sce_mean_weight_ii_4d_log_prob_expanded",
"test_sce_mean_weight_ii_expanded",
"test_sce_mean_weight_ii_log_prob",
"test_sce_mean_weight_ii_log_prob_expanded",
"test_sce_mean_weight_log_prob",
"test_sce_mean_weight_log_prob_expanded",
"test_sce_none",
"test_sce_none_expanded",
"test_sce_none_log_prob",
"test_sce_none_log_prob_expanded",
"test_sce_none_weights",
"test_sce_none_weights_expanded",
"test_sce_none_weights_log_prob",
"test_sce_none_weights_log_prob_expanded",
"test_sce_sum",
"test_sce_sum_expanded",
"test_sce_sum_log_prob",
"test_sce_sum_log_prob_expanded",
"test_nllloss_NC",
"test_nllloss_NCd1",
"test_nllloss_NCd1_ii",
"test_nllloss_NCd1_ii_expanded",
"test_nllloss_NCd1_mean_weight_negative_ii",
"test_nllloss_NCd1_mean_weight_negative_ii_expanded",
"test_nllloss_NCd1_weight",
"test_nllloss_NCd1_weight_ii",
"test_nllloss_NCd1_weight_ii_expanded",
"test_nllloss_NCd1d2",
"test_nllloss_NCd1d2_no_weight_reduction_mean_ii",
"test_nllloss_NCd1d2_no_weight_reduction_mean_ii_expanded",
"test_nllloss_NCd1d2_reduction_mean",
"test_nllloss_NCd1d2_reduction_mean_expanded",
"test_nllloss_NCd1d2_reduction_sum",
"test_nllloss_NCd1d2_with_weight",
"test_nllloss_NCd1d2_with_weight_reduction_mean",
"test_nllloss_NCd1d2_with_weight_reduction_sum",
"test_nllloss_NCd1d2_with_weight_reduction_sum_expanded",
"test_nllloss_NCd1d2_with_weight_reduction_sum_ii",
"test_nllloss_NCd1d2_with_weight_reduction_sum_ii_expanded",
"test_nllloss_NCd1d2d3_none_no_weight_negative_ii",
"test_nllloss_NCd1d2d3_none_no_weight_negative_ii_expanded",
"test_nllloss_NCd1d2d3_sum_weight_high_ii",
"test_nllloss_NCd1d2d3_sum_weight_high_ii_expanded",
"test_nllloss_NCd1d2d3d4d5_mean_weight",
"test_nllloss_NCd1d2d3d4d5_none_no_weight",
"test_center_crop_pad_crop",
"test_center_crop_pad_crop_and_pad",
"test_center_crop_pad_crop_and_pad_expanded",
"test_center_crop_pad_crop_axes_chw",
"test_center_crop_pad_crop_axes_chw_expanded",
"test_center_crop_pad_crop_axes_hwc",
"test_center_crop_pad_crop_axes_hwc_expanded",
"test_center_crop_pad_crop_expanded",
"test_center_crop_pad_crop_negative_axes_hwc",
"test_center_crop_pad_crop_negative_axes_hwc_expanded",
"test_center_crop_pad_pad",
"test_center_crop_pad_pad_expanded",
"test_gridsample_bilinear_align_corners_0_additional_1",
"test_gridsample_bilinear_align_corners_1_additional_1",
"test_gridsample_nearest_align_corners_0_additional_1",
"test_gridsample_nearest_align_corners_1_additional_1",
"test_gridsample_volumetric_bilinear_align_corners_0",
"test_gridsample_volumetric_bilinear_align_corners_1",
"test_gridsample_volumetric_nearest_align_corners_0",
"test_gridsample_volumetric_nearest_align_corners_1",
"test_onehot_negative_indices",
"test_onehot_with_axis",
"test_onehot_with_negative_axis",
"test_onehot_without_axis",
"test_dft",
"test_dft_axis_opset19",
"test_dft_inverse",
"test_dft_inverse_opset19",
"test_dft_opset19",
"test_affine_grid_3d",
"test_affine_grid_3d_align_corners",
"test_affine_grid_2d",
"test_affine_grid_2d_align_corners",
"test_rotary_embedding", //type mismatch
"test_rotary_embedding_3d_input",
"test_rotary_embedding_3d_input_expanded",
"test_rotary_embedding_expanded",
"test_rotary_embedding_interleaved",
"test_rotary_embedding_interleaved_expanded",
"test_rotary_embedding_no_position_ids",
"test_rotary_embedding_no_position_ids_expanded",
"test_rotary_embedding_no_position_ids_interleaved",
"test_rotary_embedding_no_position_ids_interleaved_expanded",
"test_rotary_embedding_no_position_ids_rotary_dim",
"test_rotary_embedding_no_position_ids_rotary_dim_expanded",
"test_rotary_embedding_with_interleaved_rotary_dim",
"test_rotary_embedding_with_interleaved_rotary_dim_expanded",
"test_rotary_embedding_with_rotary_dim",
"test_rotary_embedding_with_rotary_dim_expanded",
"test_attention_3d",
"test_attention_3d_attn_mask",
"test_attention_3d_causal",
"test_attention_3d_diff_heads_sizes",
"test_attention_3d_diff_heads_sizes_attn_mask",
"test_attention_3d_diff_heads_sizes_causal",
"test_attention_3d_diff_heads_sizes_softcap",
"test_attention_3d_diff_heads_sizes_scaled",
"test_attention_3d_gqa",
"test_attention_3d_gqa_attn_mask",
"test_attention_3d_gqa_causal",
"test_attention_3d_gqa_scaled",
"test_attention_3d_gqa_softcap",
"test_attention_3d_scaled",
"test_attention_3d_softcap",
"test_attention_3d_transpose_verification",
"test_attention_4d",
"test_attention_4d_attn_mask",
"test_attention_4d_attn_mask_3d",
"test_attention_4d_attn_mask_3d_causal",
"test_attention_4d_attn_mask_4d",
"test_attention_4d_attn_mask_4d_causal",
"test_attention_4d_attn_mask_bool",
"test_attention_4d_attn_mask_bool_4d",
"test_attention_4d_causal",
"test_attention_4d_diff_heads_sizes",
"test_attention_4d_diff_heads_sizes_attn_mask",
"test_attention_4d_diff_heads_sizes_causal",
"test_attention_4d_diff_heads_sizes_scaled",
"test_attention_4d_diff_heads_sizes_softcap",
"test_attention_4d_gqa",
"test_attention_4d_gqa_attn_mask",
"test_attention_4d_gqa_causal",
"test_attention_4d_gqa_scaled",
"test_attention_4d_gqa_softcap",
"test_attention_4d_scaled",
"test_attention_4d_softcap",
"test_attention_4d_attn_mask_bool",
"test_attention_4d_attn_mask_bool_4d",
"test_rotary_embedding_with_rotary_dim_expanded",
"test_hammingwindow",
"test_hammingwindow_expanded",
"test_hammingwindow_symmetric",
"test_hammingwindow_symmetric_expanded",
"test_hannwindow",
"test_hannwindow_expanded",
"test_hannwindow_symmetric",
"test_hannwindow_symmetric_expanded",
"test_blackmanwindow",
"test_blackmanwindow_expanded",
"test_blackmanwindow_symmetric",
"test_blackmanwindow_symmetric_expanded",
"test_layer_normalization_2d_axis0",
"test_layer_normalization_2d_axis0_expanded",
"test_layer_normalization_2d_axis0_expanded_ver18",
"test_layer_normalization_2d_axis1",
"test_layer_normalization_2d_axis1_expanded",
"test_layer_normalization_2d_axis1_expanded_ver18",
"test_layer_normalization_2d_axis_negative_1",
"test_layer_normalization_2d_axis_negative_1_expanded",
"test_layer_normalization_2d_axis_negative_1_expanded_ver18",
"test_layer_normalization_2d_axis_negative_2",
"test_layer_normalization_2d_axis_negative_2_expanded",
"test_layer_normalization_2d_axis_negative_2_expanded_ver18",
"test_layer_normalization_3d_axis0_epsilon",
"test_layer_normalization_3d_axis0_epsilon_expanded",
"test_layer_normalization_3d_axis0_epsilon_expanded_ver18",
"test_layer_normalization_3d_axis1_epsilon",
"test_layer_normalization_3d_axis1_epsilon_expanded",
"test_layer_normalization_3d_axis1_epsilon_expanded_ver18",
"test_layer_normalization_3d_axis2_epsilon",
"test_layer_normalization_3d_axis2_epsilon_expanded",
"test_layer_normalization_3d_axis2_epsilon_expanded_ver18",
"test_layer_normalization_3d_axis_negative_1_epsilon",
"test_layer_normalization_3d_axis_negative_1_epsilon_expanded",
"test_layer_normalization_3d_axis_negative_1_epsilon_expanded_ver18",
"test_layer_normalization_3d_axis_negative_2_epsilon",
"test_layer_normalization_3d_axis_negative_2_epsilon_expanded",
"test_layer_normalization_3d_axis_negative_2_epsilon_expanded_ver18",
"test_layer_normalization_3d_axis_negative_3_epsilon",
"test_layer_normalization_3d_axis_negative_3_epsilon_expanded",
"test_layer_normalization_3d_axis_negative_3_epsilon_expanded_ver18",
"test_layer_normalization_4d_axis0",
"test_layer_normalization_4d_axis0_expanded",
"test_layer_normalization_4d_axis0_expanded_ver18",
"test_layer_normalization_4d_axis1",
"test_layer_normalization_4d_axis1_expanded",
"test_layer_normalization_4d_axis1_expanded_ver18",
"test_layer_normalization_4d_axis2",
"test_layer_normalization_4d_axis2_expanded",
"test_layer_normalization_4d_axis2_expanded_ver18",
"test_layer_normalization_4d_axis3",
"test_layer_normalization_4d_axis3_expanded",
"test_layer_normalization_4d_axis3_expanded_ver18",
"test_layer_normalization_4d_axis_negative_1",
"test_layer_normalization_4d_axis_negative_1_expanded",
"test_layer_normalization_4d_axis_negative_1_expanded_ver18",
"test_layer_normalization_4d_axis_negative_2",
"test_layer_normalization_4d_axis_negative_2_expanded",
"test_layer_normalization_4d_axis_negative_2_expanded_ver18",
"test_layer_normalization_4d_axis_negative_3",
"test_layer_normalization_4d_axis_negative_3_expanded",
"test_layer_normalization_4d_axis_negative_3_expanded_ver18",
"test_layer_normalization_4d_axis_negative_4",
"test_layer_normalization_4d_axis_negative_4_expanded",
"test_layer_normalization_4d_axis_negative_4_expanded_ver18",
"test_layer_normalization_default_axis",
"test_layer_normalization_default_axis_expanded",
"test_layer_normalization_default_axis_expanded_ver18",
"test_roialign_aligned_false",
"test_roialign_aligned_true",
"test_roialign_mode_max",
"test_batchnorm_example",
"test_batchnorm_epsilon",
"test_gru_batchwise",
"test_gru_defaults",
"test_gru_seq_length",
"test_gru_with_initial_bias",
"test_reduce_l1_empty_set",
"test_reduce_l1_empty_set_expanded",
"test_reduce_l2_empty_set",
"test_reduce_l2_empty_set_expanded",
"test_reduce_log_sum_empty_set",
"test_reduce_log_sum_empty_set_expanded",
"test_reduce_log_sum_exp_empty_set",
"test_reduce_max_empty_set",
"test_reduce_min_empty_set",
"test_reduce_prod_empty_set",
"test_reduce_sum_empty_set",
"test_reduce_sum_empty_set_non_reduced_axis_zero",
"test_reduce_sum_square_empty_set",
"test_reduce_sum_square_empty_set_expanded",
"test_reduce_log_sum_exp_empty_set_expanded",
"test_loop11",
"test_eyelike_populate_off_main_diagonal",
"test_eyelike_with_dtype",
"test_eyelike_without_dtype",
"test_qlinearconv",
"test_qlinearmatmul_2D",
"test_qlinearmatmul_3D",
"test_convtranspose",
"test_convtranspose_1d",
"test_convtranspose_3d",
"test_convtranspose_dilations",
"test_convtranspose_group_2",
"test_convtranspose_group_2_image_3",
"test_convtranspose_kernel_shape",
"test_convtranspose_output_shape",
"test_convtranspose_pad",
"test_convtranspose_pads",
"test_convtranspose_with_kernel",
"test_maxpool_3d_dilations",
"test_maxpool_3d_dilations_use_ref_impl",
"test_maxpool_3d_dilations_use_ref_impl_large",
