// This file is part of OpenCV project.
// It is subject to the license terms in the LICENSE file found in the top-level directory
// of this distribution and at http://opencv.org/license.html.

#ifndef OPENCV_NDSRVP_FEATURES2D_HPP
#define OPENCV_NDSRVP_FEATURES2D_HPP

#endif
